# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Parameter parser module.

A parser is able to transform a serialized parameter value to a python object.

A parser is a callable object which takes in parameter :

- value: the serialized value.
- conf: the configuration.
- logger: a logger.
"""

__all__ = [
    'boolparser', 'pathparser', 'jsonparser', 'dictparser', 'arrayparser',
    'exprparser'
]


from b3j0f.utils.path import lookup
from b3j0f.utils.runtime import safe_eval

from json import loads as jsonloads

from re import compile as re_compile

from .parameter import Parameter
from .configuration import Configuration


def simpleparser(function):
    """Parser to use with a function which takes in parameter only one param.

    For example, simpleparser(int) returns a parser of integer.
    """

    def parse(value, *args, **kwargs):

        return function(value)

    return parse


intparser = simpleparser(int)  #: Integer value parser.
floatparser = simpleparser(float)  #: Float value parser.
complexparser = simpleparser(complex)  #: complex value parser.
strparser = simpleparser(str)  #: str value parser.


def boolparser(value, *args, **kwargs):
    """Boolean value parser.

    :param str value: value to parse.
    :return: True if value in [True, true, 1]. False Otherwise.
    :rtype: bool
    """

    return value == 'True' or value == 'true' or value == '1'


def pathparser(value, *args, **kwargs):
    """Python class path value parser.

    :param str value: python class path to parse.
    :return: lookup(value).
    """

    return lookup(value)


def jsonparser(value, *args, **kwargs):
    """Get a data from a json data format.

    :param str value: json format to parse.
    :return: data.
    :rtype: str, list, dict, int, float or bool
    """

    return jsonloads(value, *args, **kwargs)


def _typedjsonparser(value, cls, *args, **kwargs):
    """Private static method which uses the json method and check if result
    inherits from the cls. Otherwise, raise an error.

    :param str value: value to parse in a json format.
    :param type cls: expected result class.
    :return: parsed value.
    :rtype: cls
    """

    result = jsonparser(value, *args, **kwargs)

    if not isinstance(result, cls):
        raise Parameter.Error(
            'Wrong type: {0}. {1} expected.'.format(value, cls)
        )

    return result


def dictparser(value, *args, **kwargs):
    """Get a dict from a dict json format.

    :param str value: dictionary json format to parse.
    :return: parsed dictionary.
    :rtype: dict
    :raises: Parameter.Error if value is not a dict json format.
    """

    return _typedjsonparser(value, dict, *args, **kwargs)


def arrayparser(value, *args, **kwargs):
    """Get an array from:

    - an array json format.
    - a list of item name separated by commas.

    :param str value: list of items to parse. The format must be of

        - array json type.
        - named item separated by commas.

    :return: parsed array.
    :rtype: list
    :raises: Parameter.Error if value is not a list json format.
    """

    if value[0] == '[':
        result = _typedjsonparser(value, list, *args, **kwargs)

    else:
        result = list(value.split(','))

    return result

EVAL_REGEX = r'((?<=\$)[a-zA-Z_]\w*)(\.[a-zA-Z_]\w*)?'  #: eval regex.
EVAL_COMPILED_REGEX = re_compile(EVAL_REGEX)  #: eval compiled regex.


def exprparser(value, conf, *args, **kwargs):
    """Expression parser.

    Evaluate safely input value.

    Value is a lambda body expression evaluated in a safely scope (only builtin
    is loaded) where configuration categories/parameters are prefixed by the
    character '$'.

    For example, ``$cat.param`` designates the value of the category named
    ``cat`` and the parameter named ``param``. And ``$param`` designates
    the final value of the parameter named ``param``.
    """

    result = None

    def repl(match):
        """Replace matching expressin in match with corresponding conf accessor
        ."""

        result = match

        if match[1]:
            result = 'conf[{0}][{1}].value'.format(
                Configuration.VALUES, match[0]
            )

        else:
            result = 'conf[{0}][{1}].value'.format(*match)

        return result

    compilation = EVAL_COMPILED_REGEX.sub(repl, value)

    if compilation:

        _locals = {'conf': conf}

        result = safe_eval(compilation, None, _locals)

    return result



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

__all__ = ['INIConfDriver']


from os.path import join

from sys import prefix as sys_prefix

from b3j0f.utils.version import PY2

if PY2:
    from ConfigParser import (
        RawConfigParser, DuplicateSectionError, MissingSectionHeaderError
    )
else:
    from configparser import (
        RawConfigParser, DuplicateSectionError, MissingSectionHeaderError
    )

from .core import FileConfDriver


class INIConfDriver(FileConfDriver):
    """Manage ini configuration.
    """

    __register__ = True  #: Register it automatically among global managers.

    def _has_category(self, conf_resource, category, *args, **kwargs):

        return conf_resource.has_section(category.name)

    def _has_parameter(self, conf_resource, category, param, *args, **kwargs):

        return conf_resource.has_option(category.name, param.name)

    def _get_conf_resource(self, logger, conf_path=None, *args, **kwargs):

        result = RawConfigParser()

        if conf_path is not None:

            files = []

            path = FileConfDriver.get_path(conf_path)

            try:
                files = result.read(path)

            except MissingSectionHeaderError:
                logger.warning('Missing section header')

            if not files:
                result = None

        return result

    def _get_categories(self, conf_resource, *args, **kwargs):

        return conf_resource.sections()

    def _get_pnames(self, conf_resource, category, *args, **kwargs):

        return conf_resource.options(category.name)

    def _get_value(self, conf_resource, category, param, *args, **kwargs):

        return conf_resource.get(category.name, param.name)

    def _set_category(self, conf_resource, category, logger, *args, **kwargs):

        try:
            conf_resource.add_section(category.name)
        except (DuplicateSectionError, ValueError):
            pass

    def _set_parameter(self, conf_resource, category, param, *args, **kwargs):

        conf_resource.set(category.name, param.name, param.value)

    def _update_conf_resource(self, conf_resource, conf_path, *args, **kwargs):

        with open(join(sys_prefix, 'etc', conf_path), 'w') as fso:
            conf_resource.write(fso)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

__all__ = ['MetaConfDriver', 'ConfDriver']


from b3j0f.utils.version import basestring
from b3j0f.utils.path import lookup, getpath

from parser import ParserError

from ..model.configuration import Configuration
from ..model.category import Category


class MetaConfDriver(type):
    """ConfDriver meta class which register all driver in a global
    set of drivers.
    """

    def __init__(self, name, bases, attrs):

        super(MetaConfDriver, self).__init__(name, bases, attrs)

        # if the class claims to be registered
        if self.__register__:
            # add it among drivers
            ConfDriver._MANAGERS[getpath(self)] = self


class ConfDriver(object):
    """Base class for managing conf."""

    """Apply meta class for registering automatically it among global drivers
    if __register__ is True.
    """
    __metaclass__ = MetaConfDriver

    """Static param which allows this class to be automatically registered
    among drivers.
    """
    __register__ = False

    CONF_FILE = 'CONF_FILE'

    _MANAGERS = {}  #: Private set of shared driver types.

    def handle(self, conf_path, logger):
        """True iif input conf_path can be handled by self.

        :return: True iif input conf_path can be handled by self.
        :rtype: bool
        """

        conf_resource = self._get_conf_resource(
            conf_path=conf_path, logger=logger
        )

        result = conf_resource is not None

        return result

    def exists(self, conf_path):
        """True if conf_path exist related to this driver behaviour."""

        raise NotImplementedError()

    def get_conf(
            self, conf_path, logger, conf=None, override=True
    ):
        """Parse a configuration_files with input conf and returns
        parameters and errors by param name.

        :param str conf_path: conf file to parse and from get parameters.
        :param Configuration conf: conf to fill with conf_path values and
            conf param names.
        :param Logger logger: logger to use in order to trace
            information/error.
        :param bool override: if True (by default), override self
            configuration.
        """

        conf_resource = None

        result = None
        # ensure conf_path exists and is not empty.
        if self.exists(conf_path):
            try:
                # first, read conf file
                conf_resource = self._get_conf_resource(
                    conf_path=conf_path,
                    logger=logger
                )

            except Exception as ex:
                # if an error occured, log it
                logger.error(
                    'Impossible to parse conf_path {0} with {1}: {2}'
                    .format(conf_path, type(self), ex)
                )

            else:  # else process conf file
                if conf_resource is None:
                    return result

                result = Configuration() if conf is None else conf

                categories = self._get_categories(
                    conf_resource=conf_resource,
                    logger=logger
                )

                for category_name in categories:
                    # do something only for referenced categories
                    if category_name in result:

                        category = result.setdefault(
                            category_name,
                            Category(category_name)
                        )

                        pnames = self._get_pnames(
                            conf_resource=conf_resource,
                            category=category,
                            logger=logger
                        )

                        for param in category:

                            pname = param.name

                            # list of matching conf parameters
                            cparams = []

                            if isinstance(pname, basestring):

                                if pname in pnames:

                                    cparams = [category[pname].copy()]

                            else:

                                regex = pname

                                for pname in pnames:

                                    if regex.match(pname):

                                        param = category[pname].copy()
                                        param.name = pname
                                        cparams.append(param)

                            for cparam in cparams:
                                # update result
                                category.setdefault(cparam.name, cparam)
                                # get param value

                                value = self._get_value(
                                    conf_resource=conf_resource,
                                    category=category,
                                    param=cparam,
                                    logger=logger
                                )

                                # set value to param
                                if value not in (None, ''):
                                    if override or cparam.value in (None, ''):
                                        try:  # set serialized value
                                            cparam.svalue = value
                                        except ParserError as ex:
                                            pass

                                configurable = cparam.value

                                # apply conf if given
                                # get conf category name
                                conf_name = cparam.conf_name
                                confcategory = conf.get_unified_category(
                                    name=conf_name
                                )

                                if self._has_category(
                                    conf_resource=conf_resource,
                                    category=confcategory,
                                    logger=logger
                                ):

                                    conf = configurable.conf
                                    conf += confcategory

                                    conf_paths = list(configurable.conf_paths)
                                    if conf_path not in conf_paths:
                                        conf_paths.append(conf_path)

                                        configurable.conf_paths = conf_paths
                                    # apply configuration
                                    configurable.apply_configurable()

        return result

    def set_conf(self, conf_path, conf, logger):
        """Set input conf in input conf_path.

        :param str conf_path: conf file to parse and from get parameters.
        :param Configuration conf: conf to write to conf_path.
        :param Logger logger: used to log info/errors
        """

        result = None
        conf_resource = None

        try:  # get conf_resource
            conf_resource = self._get_conf_resource(
                conf_path=conf_path,
                logger=logger
            )

        except Exception as ex:
            # if an error occured, stop processing
            logger.error(
                'Impossible to parse conf_path {0}: {1}'.format(conf_path, ex)
            )
            result = ex

        # if conf_path can not be loaded, get default config conf_resource
        if conf_resource is None:
            conf_resource = self._get_conf_resource(logger=logger)

        # iterate on all conf items
        for category in conf:

            # set category
            self._set_category(
                conf_resource=conf_resource, category=category, logger=logger
            )

            # iterate on parameters
            for param in category:

                try:
                    pvalue = param.value

                except:
                    pass

                else:

                    if pvalue is not None:

                        # set param
                        self._set_parameter(
                            conf_resource=conf_resource,
                            category=category,
                            param=param,
                            logger=logger
                        )

        # write conf_resource in conf file
        self._update_conf_resource(
            conf_resource=conf_resource, conf_path=conf_path, logger=logger
        )

        return result

    @staticmethod
    def get_drivers():
        """Get global defined drivers.
        """

        return set(ConfDriver._MANAGERS.values())

    @staticmethod
    def get_driver(path):
        """Add a conf driver by its path definition.

        :param str path: driver path to add. Must be a full path from a known
            package/module.
        """

        # try to get if from global definition
        result = ConfDriver._MANAGERS.get(path)

        # if not already added
        if result is None:
            # resolve it and add it in global definition
            result = lookup(path)
            ConfDriver._MANAGERS[path] = result

        return result

    def _get_categories(self, conf_resource, logger):
        """Get a list of category names in conf_resource.

        :param conf_resource: configuration resource.
        :param Logger logger: logger to use.
        :rtype: list
        """

        raise NotImplementedError()

    def _get_pnames(self, conf_resource, category, logger):
        """Get a list of param names in conf_resource related to category.

        :param conf_resource: configuration resource.
        :param Category category: category from which get parameter names.
        :param Logger logger: logger to use.
        :rtype: list
        """

        raise NotImplementedError()

    def _has_category(self, conf_resource, category, logger):
        """True iif input conf_resource contains input category.

        :param conf_resource: configuration resource.
        :param Category category: category to check in conf_resource.
        :param Logger logger: logger to use.
        :rtype: bool
        """

        raise NotImplementedError()

    def _has_parameter(self, conf_resource, category, param, logger):
        """True iif input conf_resource has input parameter_name in input
            category.

        :param conf_resource: configuration resource.
        :param Category category: category to check in conf_resource.
        :param Parameter param: parameter to check in conf_resource.
        :param Logger logger: logger to use.
        :rtype: bool
        """

        raise NotImplementedError()

    def _get_conf_resource(self, logger, conf_path=None):
        """Get config conf_resource.

        :param str conf_path: if not None, the config conf_resource is
            conf_path content.
        :param Logger logger: logger used to log processing information
        :return: empty config conf_resource if conf_path is None, else
            conf_path content.
        """

        raise NotImplementedError()

    def _get_value(self, conf_resource, category, param, logger):
        """Get a param related to input conf_resource, category and param.

        :param conf_resource: configuration resource.
        :param Category category: category to retrieve from conf_resource.
        :param Parameter param: parameter to use from conf_resource.
        :param Logger logger: logger to use.
        """

        raise NotImplementedError()

    def _set_category(self, conf_resource, category, logger):
        """Set category on conf_resource.

        :param conf_resource: configuration resource.
        :param Category category: category to set in conf_resource.
        :param Logger logger: logger to use.
        """

        raise NotImplementedError()

    def _set_parameter(self, conf_resource, category, param, logger):
        """Set param on conf_resource.

        :param conf_resource: configuration resource.
        :param Category category: category to set in conf_resource.
        :param Parameter param: parameter to set in conf_resource.
        :param Logger logger: logger to use.
        """

        raise NotImplementedError()

    def _update_conf_resource(self, conf_resource, conf_path, logger):
        """Write conf_resource into conf_path.

        :param conf_resource: configuration resource.
        :param str conf_path: configuration path to use.
        :param Logger logger: logger to use.
        """

        raise NotImplementedError()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------


from unittest import main

from b3j0f.utils.ut import UTCase

from ..core import Configurable
from ...model.configuration import Configuration
from ...model.category import Category
from ...model.parameter import Parameter
from ...model.parser import intparser, floatparser

from tempfile import NamedTemporaryFile

from os import remove


class ConfigurableTest(UTCase):

    def setUp(self):

        self.conf_paths = (
            NamedTemporaryFile().name,
            NamedTemporaryFile().name
        )

        self.configurable = Configurable()

        self.conf = Configuration(
            Category(
                'A',
                Parameter('a', value='a'),
                Parameter('_', value=2, parser=intparser),
                Parameter('error', parser=floatparser, svalue='error')
            ),
            Category(
                'B',
                Parameter('a', value='b'),
                Parameter('b', value='b')
            )
        )

    def test_configuration_files(self):

        configurable = Configurable()
        configurable.conf_paths = self.conf_paths

        self.assertEqual(
            configurable.conf_paths,
            self.conf_paths)

        configurable = Configurable(conf_paths=self.conf_paths)

        self.assertEqual(configurable.conf_paths, self.conf_paths)

    def test_auto_conf(self):

        configurable = Configurable()

        self.assertTrue(configurable.auto_conf)

        configurable.auto_conf = False

        self.assertFalse(configurable.auto_conf)

    def test_logging_level(self):

        configurable = Configurable()

        self.assertTrue(configurable.log_lvl, 'INFO')

        configurable = Configurable(log_lvl='DEBUG')

        self.assertTrue(configurable.log_lvl, 'DEBUG')

        configurable.log_lvl = 'INFO'

        self.assertTrue(configurable.log_lvl, 'INFO')

    def test_configuration(self):

        # test to get from no file
        configurable = Configurable()

        conf = configurable.get_conf()

        self.assertEqual(len(conf), len(self.conf))

        # test to get from files which do not exist
        configurable.conf_paths = self.conf_paths

        for conf_path in self.conf_paths:
            try:
                remove(conf_path)
            except OSError:
                pass

        conf = configurable.get_conf()

        self.assertEqual(len(conf), len(self.conf))

        # get parameters from empty files
        for conf_path in self.conf_paths:
            with open(conf_path, 'w') as _file:
                _file.write('\n')

        conf = configurable.get_conf()

        self.assertEqual(len(conf), len(self.conf))

        # get parameters from empty files and empty parsing_rules
        conf = Configuration()
        configurable.get_conf(conf=conf)

        self.assertEqual(len(conf), 0)

        # fill files
        configurable = Configurable(conf_paths=self.conf_paths)

        # add first category in conf file[0]
        configurable.set_conf(
            conf_path=self.conf_paths[0],
            conf=Configuration(self.conf['A']),
            driver=configurable._drivers.split(',')[0]
        )

        # add second category in conf file[1]
        configurable.set_conf(
            conf_path=self.conf_paths[1],
            conf=Configuration(self.conf['B']),
            driver=configurable._drivers.split(',')[1]
        )

        conf = configurable.get_conf(conf=self.conf)

        unified_configuration = conf.unify()
        parameters = unified_configuration[Configuration.VALUES]
        errors = unified_configuration[Configuration.ERRORS]

        self.assertIn('a', parameters)
        self.assertNotIn('a', errors)
        self.assertEqual(parameters['a'].value, 'b')
        self.assertIn('_', parameters)
        self.assertNotIn('_', errors)
        self.assertEqual(parameters['_'].value, 2)
        self.assertIn('b', parameters)
        self.assertNotIn('b', errors)
        self.assertEqual(parameters['b'].value, 'b')
        self.assertIn('error', errors)
        self.assertNotIn('error', parameters)

    def test_reconfigure(self):

        self.assertTrue(self.configurable.auto_conf)

        conf = Configuration(
            Category(
                'TEST',
                Parameter('auto_conf', value=False)))

        self.configurable.configure(conf=conf)
        self.assertFalse(self.configurable.auto_conf)
        self.assertEqual(self.configurable.log_lvl, 'INFO')

        conf = Configuration(
            Category('TEST', Parameter('log_lvl', value='DEBUG'))
        )

        self.configurable.configure(conf=conf)
        self.assertEqual(self.configurable.log_lvl, 'INFO')

        self.configurable.reconf_once = True
        self.configurable.configure(conf=conf)
        self.assertEqual(self.configurable.log_lvl, 'DEBUG')
        self.assertFalse(self.configurable.reconf_once)

        self.configurable.log_lvl = 'INFO'
        self.configurable.auto_conf = True
        self.configurable.configure(conf=conf)
        self.assertEqual(self.configurable.log_lvl, 'DEBUG')

    def test_configure_to_reconfigure_param(self):
        """Test to reconfigure an object with to_configure parameter.
        """

        self.assertTrue(self.configurable.auto_conf)

        class ToConfigure(object):
            """Class to configure.
            """
            def __init__(self):
                super(ToConfigure, self).__init__()
                self.test = None

        to_configure = ToConfigure()

        param = 'test'

        conf = Configuration(
            Category(
                'TEST',
                Parameter(param, value=True)
            )
        )

        self.configurable.configure(conf=conf, to_configure=to_configure)
        self.assertTrue(to_configure.test)

    def test_configure_without_inheritance(self):
        """Test to configure an object without inheritance.
        """

        class ToConfigure(object):
            """Class to configure.
            """
            def __init__(self):

                super(ToConfigure, self).__init__()

                self.test = None

        to_configure = ToConfigure()

        configurable = Configurable(to_configure=to_configure)

        param = 'test'

        conf = Configuration(
            Category('TEST', Parameter(param, value=True))
        )

        configurable.configure(conf=conf, to_configure=to_configure)
        self.assertTrue(to_configure.test)

    def test_parser_inheritance(self):

        class _Configurable(Configurable):

            def _conf(self, *args, **kwargs):

                result = super(_Configurable, self)._conf(*args, **kwargs)

                result += Category('PLOP')

                return result

        configurable = Configurable()

        _configurable = _Configurable()

        self.assertEqual(
            len(configurable.conf) + 1,
            len(_configurable.conf)
        )

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Configuration definition objects."""

__all__ = ['Configuration']


from b3j0f.utils.version import OrderedDict

from .category import Category


class Configuration(object):
    """Manage conf such as a list of Categories.

    The order of categories permit to ensure param overriding.
    """

    ERRORS = '_ERRORS'  #: category name which contains errors.
    VALUES = '_VALUES'  #: category name which contains local param values.
    FOREIGNS = '_FOREIGN'  #: category name which contains foreign params vals.

    def __init__(self, *categories):
        """
        :param list categories: categories to configure.
        """

        super(Configuration, self).__init__()

        self.categories = OrderedDict()

        # set categories
        for category in categories:
            self.categories[category.name] = category

    def __iter__(self):

        return iter(self.categories.values())

    def __delitem__(self, category_name):

        del self.categories[category_name]

    def __getitem__(self, category_name):

        return self.categories[category_name]

    def __contains__(self, category_name):

        return category_name in self.categories

    def __len__(self):

        return len(self.categories)

    def __iadd__(self, other):
        """Add categories or conf categories in self."""

        # if other is a conf add a copy of all other categories
        if isinstance(other, Configuration):

            for category in other:
                self += category

        else:  # in case of category
            category = self.categories.get(other.name)

            if category is None:
                self.put(other)

            else:
                for param in other:
                    category.put(param)

        return self

    def __repr__(self):

        return 'Configuration({0})'.format(self)

    def get(self, category_name, default=None):
        """Get a category by its name.

        :param str category_name: category name.
        :param default: default value if category_name is not registered.
        :return: related Category.
        :rtype: Category
        """

        return self.categories.get(category_name, default)

    def setdefault(self, category_name, category):
        """Register a category with specific name if name not already used.

        :param str category_name: category name.
        :param Category category: to register.
        :return: existing category or input category if no category have been
            registered with this name.
        :rtype: Category
        """

        return self.categories.setdefault(category_name, category)

    def put(self, category):
        """Put a category and return the previous one if exist."""

        result = self.categories.get(category.name)
        self.categories[category.name] = category

        return result

    def unify(self, copy=False):
        """Get a conf which contains only two categories:

        - VALUES where params are all self params where values are not
            exceptions.contains all values
        - ERRORS where params are all self params where values are
            exceptions

        :param bool copy: copy self params (default False).
        :return: two categories named respectivelly VALUES and ERRORS and
            contain respectivelly self param values and parsing errors.
        :rtype: Configuration
        """

        result = Configuration()

        values = Category(Configuration.VALUES)
        errors = Category(Configuration.ERRORS)
        foreigns = Category(Configuration.FOREIGNS)

        for category in self:

            for param in category:

                pvalue = None

                try:  # get value
                    pvalue = param.value

                except:
                    pass

                if (pvalue, param.error) != (None, None):
                    final_values = values if param.local else foreigns

                    if param.error is not None:  # if param is in error
                        to_update = errors
                        to_delete = final_values

                    else:
                        to_update = final_values
                        to_delete = errors

                    to_update.put(param.copy() if copy else param)

                    if param.name in to_delete:
                        del to_delete[param.name]

        result += values
        result += foreigns
        result += errors

        return result

    def get_unified_category(self, name, copy=False):
        """Add a category with input name which takes all params provided
        by other categories.

        :param str name: new category name.

        :param bool copy: copy self params (default False).
        """

        result = Category(name)

        for category in self.categories.values():
            for param in category.values():
                result.put(param.copy() if copy else param)

        return result

    def add_unified_category(self, name=VALUES, copy=False, new_content=None):
        """Add a unified category to self and add new_content if not None."""

        category = self.get_unified_category(name=name, copy=copy)

        if new_content is not None:
            category += new_content

        self += category

    def add_param_list(self, name, content):
        """Add a list of parameters to self."""

        self.categories[name] = content

    def clean(self):
        """Clean this params in setting value to None."""

        for category in self.categories:

            category.clean()

    def copy(self, cleaned=False):
        """Copy this Configuration.

        :param bool cleaned: copy this element without parameter values.
        """

        result = Configuration()

        for category in self.categories.values():
            result.put(category.copy(cleaned=cleaned))

        return result

    def update(self, conf):
        """Update this content with input conf."""

        for category in conf:
            category = self.categories.setdefault(
                category.name, category.copy()
            )

            for param in category:
                param = category.setdefault(
                    param.name, param.copy()
                )



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Configuration definition objects."""

__all__ = ['Parameter']


from b3j0f.utils.version import basestring

from re import compile as re_compile


class Parameter(object):
    """Parameter identified among a category by its name.

    Provide a value (None by default) and a parser (str by default).

    Properties are:

    - name: parameter name. If regex, drivers will generated as many parameters
        as they exist in the configuration resource with this properties (
        value, conf, parser, etc.).
    - value: parameter value. None if error is not None.
    - error: parameter error encountered while attempting to set values.
    - parser: parameter value parser from configuration resources.
    - conf: parameter configuration if not None. If not None, the value must be
        a callable beceause the conf is used such as a kwargs.
    - critical: if True (False by default), set a configurable property
    - local: distinguish local parameters from those found in configuration
        resources.
    - asitem: distinguish parameters from those in a ParamList.
    """

    CONF_SUFFIX = '::conf'  #: parameter configuration name.

    PARAM_NAME_REGEX = '[a-zA-Z_][a-zA-Z0-9_]*'  #: simple name validation.
    #: simple name regex compiler.
    _PARAM_NAME_COMPILER = re_compile(PARAM_NAME_REGEX)

    class Error(Exception):
        """Handle Parameter errors."""

    def __init__(
            self, name, vtype=object, parser=None, svalue=None, value=None,
            conf=None, critical=False, local=True, asitem=None
    ):
        """
        :param name: depends on type:
            - str: unique by category.
            - regex: designates a group of parameters with a matching name and
                common properties (parser, value, conf, etc.).
        :param type vtype: parameter value type.
        :param callable parser: param test deserializer which takes in param
            a str.
        :param str svalue: serialized value.
        :param value: param value. None if not given.
        :param dict conf: parameter configuration used to init the parameter
            value if not None. In this case, the parameter value must be
            callable.
        :param bool critical: True if this parameter is critical. A critical
            parameter can require to restart a component for example.
        :param bool local: distinguish local parameters from those found in
            configuration resources.
        :param Category asitem: distinguish parameters from those in a
            ParamList.
        """

        super(Parameter, self).__init__()

        # init private attributes
        self._name = None
        self._value = None
        self._error = None
        self._svalue = None

        # init public attributes
        self.name = name
        self.vtype = vtype
        self.conf = conf
        self.parser = parser
        self.critical = critical
        self.local = local
        self.asitem = asitem
        self.svalue = svalue
        self.value = value

    def __eq__(self, other):

        return isinstance(other, Parameter) and other.name == self.name

    def __hash__(self):

        return hash(self.name)

    def __repr__(self):

        return 'Parameter({0}, {1}, {2})'.format(
            self.name, self.value, self.parser
        )

    @property
    def error(self):
        """Get encountered error."""

        return self._error

    @property
    def name(self):
        """Get parameter name.

        :return: this name.
        :rtype: str
        """

        return self._name

    @name.setter
    def name(self, value):
        """Set parameter name.

        :param str value: name value.
        """

        if isinstance(value, basestring):

            if not Parameter._PARAM_NAME_COMPILER.match(value):
                value = re_compile(value)

        self._name = value

    @property
    def conf_name(self):
        """Get this configuration name.

        :return: {self.name}{CONF_SUFFIX}.
        :rtype: str
        """

        return '{0}{1}'.format(self.name, Parameter.CONF_SUFFIX)

    @property
    def svalue(self):
        """Get serialized value.

        :rtype: str
        """

        return self._svalue

    @svalue.setter
    def svalue(self, value):
        """Change of serialized value.

        Nonify this value as well.

        :param str value: serialized value to use.
        """

        self.value = None
        self._svalue = value

    @property
    def value(self):
        """Get parameter value.

        If this cached value is None and this serialized value is not None,
        calculate the new value from the serialized one.

        :return: parameter value.
        :raises: TypeError if serialized value is not an instance of self vtype
            . ParserError if parsing step raised an error.
        """

        # if cached value is None and serialiazed value exists
        if self._value is None and self._svalue is not None:

            self._error = None  # nonify error.

            if self.parser is None:

                self.value = self._svalue

            else:
                # parse value if str and if parser exists
                try:
                    finalvalue = self.parser(self._svalue, conf=self)

                except Exception as ex:

                    self._error = ex
                    raise Parameter.Error(
                        'Impossible to parse value "{0}"" with {1}.'.format(
                            self._svalue, self.parser
                        )
                    ).with_traceback(ex.__traceback__)

                else:
                    # try to apply conf
                    if finalvalue is None or self.conf is None:
                        self.value = finalvalue

                    else:  # apply conf
                        try:
                            self.value = finalvalue(**self.conf)
                        except Exception as ex:
                            self._error = ex
                            raise ex

        return self._value

    @value.setter
    def value(self, value):
        """Change of parameter value.

        If an error occured, it is stored in this error attribute.

        :param value: new value to use. If input value is not an instance of
            self.vtype, self error
        :raises: TypeError if input value is not an instance of self vtype.
        """

        if value is None or (
                self.vtype is not None and isinstance(value, self.vtype)
        ):

            self._value = value

        else:
            # raise wrong type error
            error = TypeError(
                'Wrong value type of {0}. {1} expected.'.format(
                    value, self.vtype
                )
            )
            self._error = error
            raise error

    def copy(self, name=None, cleaned=False):
        """Get a copy of the parameter with specified name and new value if
        cleaned.

        :param str name: new parameter name.
        :param bool cleaned: clean the value.
        :return: new parameter.
        """

        if name is None:
            name = self.name

        kwargs = {
            'name': name,
            'parser': self.parser,
            'local': self.local,
            'svalue': self.svalue,
            'critical': self.critical,
            'vtype': self.vtype,
            'conf': self.conf,
            'asitem': self.asitem,
            'value': self._value
        }

        result = Parameter(**kwargs)

        if cleaned:
            result.clean()

        return result

    def clean(self):
        """Clean this param in removing values."""

        self._value = None
        self._svalue = None



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Configuration definition objects."""

__all__ = ['Category']


from collections import Iterable

from .parameter import Parameter


class Category(dict):
    """Parameter category which contains a dictionary of params.
    """

    def __init__(self, name, *params):
        """
        :param str name: unique in a conf.
        :param list params: Parameters
        """

        super(Category, self).__init__()

        self.name = name

        for param in params:
            self[param.name] = param

    def __iter__(self):

        return iter(self.values())

    def __eq__(self, other):

        return isinstance(other, Category) and other.name == self.name

    def __hash__(self):

        return hash(self.name)

    def __repr__(self):

        return 'Category({0}, {1})'.format(self.name, self.values())

    def __iadd__(self, value):

        if isinstance(value, Category):
            self += value.values()

        elif isinstance(value, Iterable):
            for content in value:
                self += content

        elif isinstance(value, Parameter):
            self.put(value)

        else:
            raise Exception(
                'Wrong value {0} to add. (Param | Category)+ expected.'
                .format(value)
            )

        return self

    def put(self, param):
        """Put a param and return the previous one if exist
        """

        result = self.get(param.name)
        self[param.name] = param

        return result

    def clean(self):
        """Clean this params in setting value to None."""

        for param in self.values():

            param.clean()

    def copy(self, name=None, cleaned=False):
        """Copy this Category.

        :param bool cleaned: copy this element without parameter values.
        """

        if name is None:
            name = self.name

        result = Category(name)

        for param in self:
            _param = param.copy(cleaned=cleaned)
            result.put(_param)

        return result
